/*
 * Copyright (c) 2008
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static const char copyright[] =
    "@(#) Copyright (c) 2008\n\
The Regents of the University of California.  All rights reserved.\n";
static const char rcsid[] =
    "@(#) $Id: bclient.c 10 2009-07-17 16:57:08Z leres $ (LBL)";
#endif

/*
 * bclient - example broccoli client
 */

#include <sys/types.h>
#include <sys/socket.h>

#include <arpa/inet.h>

#include <broccoli.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>

#include "common.h"
#include "gnuc.h"

/* Globals */
int debug;
int verbose;
const char *prog;
const char *addr = DEFAULTADDR;
const char *port = DEFAULTPORT;
const char *eventstr = EVENTSTR;
static int timeout = 10;			/* select timeout */
u_int32_t cookie;

/* Externs */
extern int optind;
extern int opterr;
extern char *optarg;

/* Forwards */
BroEvent *acldevent(const char *);
void callback(BroConn *, void *, BroRecord *);
int main(int, char **);
void loop(int, BroConn *);
__dead void usage(void) __attribute__((noreturn));

int
main(int argc, char **argv)
{
	int i, op, s;
	char *cp, *ep;
	BroConn *bc;
	struct sockaddr_in sin;

	if (argv[0] == NULL)
		prog = "acld";
	else if ((cp = strrchr(argv[0], '/')) != NULL)
		prog = cp + 1;
	else
		prog = argv[0];

	opterr = 0;
	while ((op = getopt(argc, argv, "dDva:e:P:")) != EOF)
		switch (op) {

		case 'a':
			addr = optarg;
			break;

		case 'd':
			++debug;
			break;

		case 'D':
			++bro_debug_messages;
			break;

		case 'e':
			eventstr = optarg;
			break;

		case 'P':
			port = optarg;
			break;

		case 'v':
			++verbose;
			break;

		default:
			usage();
			/* NOTREACHED */
		}

	if (argc != optind)
		usage();

	memset(&sin, 0, sizeof(sin));
	sin.sin_len = sizeof(sin);
	sin.sin_family = AF_INET;

	i = strtol(port, &ep, 10);
	if (*ep != '\0' || i < 1 || i > 0xffff) {
		fprintf(stderr, "%s: bad port value (%s)\n", prog, port);
		exit(EX_USAGE);
	}
	sin.sin_port = htons(i);

	/* "*" -> 0.0.0.0 else decode string address */
	if (strcmp(addr, "*") != 0 &&
	    inet_pton(sin.sin_family, addr, &sin.sin_addr) < 0) {
		fprintf(stderr, "%s: bad address value (%s): %s\n",
		    prog, addr, strerror(errno));
		exit(EX_USAGE);
	}

	s = socket(PF_INET, SOCK_STREAM, 0);
	if (s < 0) {
		fprintf(stderr, "%s: socket: %s\n", prog, strerror(errno));
		exit(EX_OSERR);
	}

	while (connect(s, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
		fprintf(stderr, "%s: connect: %s\n", prog, strerror(errno));
		sleep(1);
	}

	bc = bro_conn_new_socket(s, BRO_CFLAG_RECONNECT);
	if (bc == NULL) {
		fprintf(stderr, "%s: bro_conn_new_socket to %s:%d failed\n",
		    prog, inet_ntoa(sin.sin_addr), ntohs(sin.sin_port));
		exit(EX_UNAVAILABLE);
	}
	fprintf(stderr, "%s: Connecting to %s:%d\n",
	    prog, inet_ntoa(sin.sin_addr), ntohs(sin.sin_port));

	bro_event_registry_add(bc, eventstr, (BroEventFunc)callback, NULL);

	while (!bro_conn_connect(bc)) {
		fprintf(stderr, "%s: Failed to connect; retrying...\n", prog);
		sleep(1);
	}

	fprintf(stderr, "%s: connected\n", prog);

	loop(s, bc);
	bro_conn_delete(bc);

	exit(0);
}

BroEvent *
acldevent(const char *buf)
{
	BroEvent *ev;
	BroRecord *rec;
	BroString str;

	rec = bro_record_new();
	if (rec == NULL) {
		fprintf(stderr,
		    "%s: bro_record_new() failed\n",
		    prog);
		return (NULL);
	}

	/* "cmd" */
	bro_string_init(&str);
	bro_string_set(&str, buf);
	if (!bro_record_add_val(rec, EV_CMD, BRO_TYPE_STRING, NULL, &str)) {
		fprintf(stderr, "%s: bro_record_add_val(%s) failed\n",
		    prog, EV_CMD);
		bro_record_free(rec);
		return (NULL);
	}
	bro_string_cleanup(&str);

	++cookie;
	if (!bro_record_add_val(rec, EV_COOKIE, BRO_TYPE_COUNT,
	    NULL, &cookie)) {
		fprintf(stderr, "%s: bro_record_add_val(%s) failed\n",
		    prog, EV_CMD);
		bro_record_free(rec);
		return (NULL);
	}

	ev = bro_event_new(eventstr);

	bro_event_add_val(ev, BRO_TYPE_RECORD, NULL, rec);
	bro_record_free(rec);
	return (ev);
}

void
callback(BroConn *bc, void *data, BroRecord *rec)
{
	char cmd[64];
	char payload[1024];
	uint32 cookie;
	struct timeval tv;

	if (!extractstr(rec, EV_CMD, cmd, sizeof(cmd))) {
		fprintf(stderr, "%s: missing %s\n", prog, EV_CMD);
		return;
	}
	if (!extractuint32(rec, EV_COOKIE, &cookie)) {
		fprintf(stderr, "%s: missing %s\n", prog, EV_COOKIE);
		return;
	}
	if (!extractlong(rec, EV_SEC, &tv.tv_sec)) {
		fprintf(stderr, "%s: missing %s\n", prog, EV_SEC);
		return;
	}
	if (!extractlong(rec, EV_USEC, &tv.tv_usec)) {
		fprintf(stderr, "%s: missing %s\n", prog, EV_USEC);
		return;
	}
	fprintf(stderr, "%s: %ld.%06ld recv:%d \"%s\"\n",
	    prog, tv.tv_sec, tv.tv_usec, cookie, cmd);

	if (extractstr(rec, EV_PAYLOAD, payload, sizeof(payload)))
		fprintf(stderr, "------\n%s------\n", payload);

}

void
loop(int s, BroConn *bc)
{
	int n, nfds;
	fd_set readfds;
	struct timeval tmo;
	BroEvent *ev;
	char buf[1024];
	ssize_t cc;

	for (;;) {
		/* Recalculate fdsets */
		FD_ZERO(&readfds);
		FD_SET(fileno(stdin), &readfds);
		nfds = fileno(stdin);
		FD_SET(s, &readfds);
		if (nfds < s)
			nfds = s;
		++nfds;

		/* Set timeout */
		tmo.tv_sec = timeout;
		tmo.tv_usec = 0;
		n = select(nfds, &readfds, NULL, NULL, &tmo);

		if (n < 0) {
			/* Don't choke if we get ptraced */
			if (errno == EINTR)
				continue;
			fprintf(stderr, "%s: select: %s\n",
			    prog, strerror(errno));
			exit(EX_OSERR);
		}

		/* Timeout */
		if (n == 0)
			bro_event_queue_flush(bc);

		/* Data from server */
		if (FD_ISSET(s, &readfds)) {
			--n;
			if (!bro_conn_process_input(bc) &&
			    !bro_conn_alive(bc)) {
				fprintf(stderr, "%s: server exited\n", prog);
				return;
			}
		}

		/* Input from stdin */
		if (FD_ISSET(fileno(stdin), &readfds)) {
			--n;
			cc = read(fileno(stdin), buf, sizeof(buf));
			if (cc < 0)
				fprintf(stderr, "%s: read stdin: %s\n",
				    prog, strerror(errno));
			else if (cc == 0) {
				fprintf(stderr, "%s: EOF on stdin\n", prog);
				break;
			} else {
				if (cc >= sizeof(buf))
					cc = sizeof(buf) - 1;
				if (cc > 0 && buf[cc - 1] == '\n')
					--cc;
				buf[cc] = '\0';
				ev = acldevent(buf);

				if (!bro_event_send(bc, ev))
					fprintf(stderr, "%s: failed to send\n",
					    prog);
				else
					fprintf(stderr, "%s: sent:%d \"%s\"\n",
					    prog, cookie, buf);
				bro_event_free(ev);
			}
		}

		if (n != 0)
			fprintf(stderr, "%s: select leakage: %d\n", prog, n);
			continue;
	}
}

__dead void
usage(void)
{

	fprintf(stderr, "%s version %s\n", prog, version);
	fprintf(stderr,
	    "Usage: %s [-dDv] [-e eventstr] [-a addr] [-P port]\n", prog);
	exit(EX_USAGE);
}
