/*
 * Copyright (c) 2008
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static const char copyright[] =
    "@(#) Copyright (c) 2008\n\
The Regents of the University of California.  All rights reserved.\n";
static const char rcsid[] =
    "@(#) $Id: bserver.c 9 2008-10-25 01:56:33Z leres $ (LBL)";
#endif

/*
 * bserver - example broccoli server
 */

#include <sys/types.h>
#include <sys/socket.h>

#include <arpa/inet.h>

#include <broccoli.h>
#include <errno.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>

#include "common.h"
#include "gnuc.h"

struct client {
	struct client *next;
	u_int n;			/* client number */
	int c;				/* client socket */
	BroConn *bc;
	struct in_addr addr;
};

/* Globals */
int debug;
int verbose;
const char *prog;
const char *addr = DEFAULTADDR;
const char *port = DEFAULTPORT;
const char *eventstr = EVENTSTR;
static int backlog = 16;			/* listen() backlog */
static int timeout = 10;			/* select timeout */
struct client *clients;
static u_int nclients;				/* next client number */

/* Externs */
extern int optind;
extern int opterr;
extern char *optarg;

/* Forwards */
void callback(BroConn *, void *, BroRecord *);
void freeclient(struct client *);
void loop(int);
int main(int, char **);
struct client *newclient(int);
__dead void usage(void) __attribute__((noreturn));

int
main(int argc, char **argv)
{
	int i, op, s;
	char *cp, *ep;
	struct sockaddr_in sin;
	static int on = 1;

	if (argv[0] == NULL)
		prog = "acld";
	else if ((cp = strrchr(argv[0], '/')) != NULL)
		prog = cp + 1;
	else
		prog = argv[0];

	opterr = 0;
	while ((op = getopt(argc, argv, "dDva:e:P:")) != EOF)
		switch (op) {

		case 'a':
			addr = optarg;
			break;

		case 'd':
			++debug;
			break;

		case 'D':
			++bro_debug_messages;
			break;

		case 'e':
			eventstr = optarg;
			break;

		case 'P':
			port = optarg;
			break;

		case 'v':
			++verbose;
			break;

		default:
			usage();
			/* NOTREACHED */
		}

	if (argc != optind)
		usage();

	memset(&sin, 0, sizeof(sin));
	sin.sin_len = sizeof(sin);
	sin.sin_family = AF_INET;

	i = strtol(port, &ep, 10);
	if (*ep != '\0' || i < 1 || i > 0xffff) {
		fprintf(stderr, "%s: bad port value (%s)\n", prog, port);
		exit(EX_USAGE);
	}
	sin.sin_port = htons(i);

	/* "*" -> 0.0.0.0 else decode string address */
	if (strcmp(addr, "*") != 0 &&
	    inet_pton(sin.sin_family, addr, &sin.sin_addr) < 0) {
		fprintf(stderr, "%s: bad address value (%s): %s\n",
		    prog, addr, strerror(errno));
		exit(EX_USAGE);
	}

	s = socket(PF_INET, SOCK_STREAM, 0);
	if (s < 0) {
		fprintf(stderr, "%s: socket: %s\n", prog, strerror(errno));
		exit(EX_OSERR);
	}
	if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR,
	    (char *)&on, sizeof (on)) < 0) {
		fprintf(stderr, "%s: setsockopt: %s\n", prog, strerror(errno));
		exit(EX_OSERR);
	}
	if (bind(s, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
		fprintf(stderr, "%s: bind: %s\n", prog, strerror(errno));
		exit(EX_OSERR);
	}
	if (listen(s, backlog) < 0) {
		fprintf(stderr, "%s: listen: %s\n", prog, strerror(errno));
		exit(EX_OSERR);
	}

	loop(s);
	exit(0);
}

void
callback(BroConn *bc, void *data, BroRecord *rec)
{
	struct client *cl;
	int i;
	uint32 cookie;
	BroString bstr;
	BroEvent *ev;
	char cmd[64];
	char args[1024];
	char buf[128];

	cl = data;
	if (!extractstr(rec, EV_CMD, cmd, sizeof(cmd))) {
		fprintf(stderr, "%s: #%d recv NULL\n", prog, cl->n);
		return;
	}

	if (!extractstr(rec, EV_ARGS, args, sizeof(args)))
		args[0] = '\0';

	if (!extractuint32(rec, EV_COOKIE, &i))
		cookie = 0;
	else
		cookie = i;

	fprintf(stderr, "%s: #%d recv:%u \"%s\" \"%s\"\n",
	    prog, cl->n, cookie, cmd, args);

	ev = bro_event_new(eventstr);
	snprintf(buf, sizeof(buf), "roger-%s", cmd);
	bro_string_set(&bstr, buf);
	bro_event_add_val(ev, BRO_TYPE_STRING, NULL, &bstr);
	bro_string_cleanup(&bstr);
	if (!bro_event_send(bc, ev))
		fprintf(stderr, "%s: failed to send\n", prog);
	else
		fprintf(stderr, "%s: #%d sent \"%s\"\n", prog, cl->n, buf);
	bro_event_free(ev);
}

void
freeclient(struct client *cl)
{
	struct client *cl2;

	if (cl->c >= 0) {
		close(cl->c);
		cl->c = -1;
	}
	if (cl->bc != NULL) {
		bro_conn_delete(cl->bc);
		cl->bc = NULL;
	}

	if (clients == cl) {
		clients = clients->next;
	} else {
		for (cl2 = clients; cl2 != NULL; cl2 = cl2->next)
			if (cl2->next == cl)
				break;
		if (cl2 == NULL) {
			fprintf(stderr,
			    "%s: freeclient: client #%d: can't happen",
			    prog, cl->n);
			exit(EX_SOFTWARE);
		}
		cl2->next = cl->next;
	}
	cl->next = NULL;

	free(cl);
}

void
loop(int s)
{
	int c, n, nfds;
	fd_set readfds;
	struct timeval tmo;
	socklen_t csalen;
	struct sockaddr_in csa;
	struct client *cl, *nextcl;
	const char *fstr = "%s: #%d %s failed (%s)\n";

	for (;;) {
		/* Recalculate fdsets */
		FD_ZERO(&readfds);
		FD_SET(s, &readfds);
		nfds = s;
		for (cl = clients; cl != NULL; cl = cl->next) {
			FD_SET(cl->c, &readfds);
			nfds = MAX(nfds, cl->c);
		}
		++nfds;

		/* Set timeout */
		tmo.tv_sec = timeout;
		tmo.tv_usec = 0;
		n = select(nfds, &readfds, NULL, NULL, &tmo);
		if (n < 0) {
			/* Don't choke if we get ptraced */
			if (errno == EINTR)
				continue;
			fprintf(stderr, "%s: select: %s",
			    prog, strerror(errno));
			exit(EX_OSERR);
		}

		/* Timeout */
		if (n == 0)
			continue;

		/* New client */
		if (FD_ISSET(s, &readfds)) {
			--n;
			memset(&csa, 0, sizeof(csa));
			csalen = sizeof(csa);
			c = accept(s, (struct sockaddr *)&csa, &csalen);
			cl = newclient(c);
			cl->addr = csa.sin_addr;
			cl->bc = bro_conn_new_socket(cl->c, BRO_CFLAG_NONE);
			if (cl->bc == NULL) {
				fprintf(stderr, fstr, prog, cl->n,
				    inet_ntoa(cl->addr), "bro_conn_new_socket");
				freeclient(cl);
				cl = NULL;
			} else {
				bro_event_registry_add(cl->bc, eventstr,
				    (BroEventFunc)callback, cl);
				if (!bro_conn_connect(cl->bc)) {
					fprintf(stderr, fstr, prog, cl->n,
					    inet_ntoa(cl->addr),
					    "bro_conn_connect");
					freeclient(cl);
					cl = NULL;
				}
			}
			if (cl != NULL)
				fprintf(stderr, "%s: #%d connected (%s)\n",
				    prog, cl->n, inet_ntoa(cl->addr));
		}

		/* Service clients */
		for (cl = clients; cl != NULL; cl = nextcl) {
			/* Ok to free client inside this loop */
			nextcl = cl->next;
			if (cl->c >= 0 && FD_ISSET(cl->c, &readfds)) {
				--n;
				if (!bro_conn_process_input(cl->bc) &&
				    !bro_conn_alive(cl->bc)) {
					freeclient(cl);
					fprintf(stderr,
					    "%s: #%d exited\n",
					    prog, cl->n);
				}
			}
		}

		if (n != 0)
			fprintf(stderr, "%s: select leakage: %d\n", prog, n);
	}
}

struct client *
newclient(int c)
{
	struct client *cl, *cl2;

	cl = new(1, sizeof(*cl), "newclient cl");
	cl->c = c;
	cl->n = nclients++;

	/* Append to end of client list */
	if (clients == NULL) {
		clients = cl;
	} else {
		for (cl2 = clients; cl2 != NULL; cl2 = cl2->next)
			if (cl2->next == NULL)
				break;
		cl2->next = cl;
	}
	return (cl);
}

__dead void
usage(void)
{

	fprintf(stderr, "%s version %s\n", prog, version);
	fprintf(stderr,
	    "Usage: %s [-dDv] [-e eventstr] [-a addr] [-P port]\n", prog);
	exit(EX_USAGE);
}
